class Semantic {
  constructor(matcher, action, uuid) {
    this.matcher = matcher;
    this.action = action;
    this.uuid = uuid;
  }

  toString() {
    return `Semantic(${this.matcher}, ${this.action})`;
  }

  getAPI(config) {
    if (config && config.getAPI) {
      return config.getAPI(this.uuid)
    }
  }

  matches(objects, context, config) {
    return this.matcher({objects, global: objects, context, api: this.getAPI(config)})
  }

  apply(objects, context, config, response, s) {
    const context_prime = Object.assign({}, context);
    let n = (id) => id;
    if (config && 'nsToString' in config) {
      n = (id) => config.nsToString(id);
    }
    this.action({objects, global: objects, n, context: context_prime, uuid: this.uuid, config, response, s, api: this.getAPI(config)});
    return context_prime;
  }
}

class Semantics {
  constructor(semantics) {
    this.semantics = semantics
    this.logs = []
  };

  applyToContext(objects, context, config, response) {
    let context_prime = {}
    if (! (context instanceof Array || context instanceof Object) ){
      return context;
    }

    for ( const key in context ) {
      context_prime[key] = this.apply(objects, context[key], config, response)
    }
    const s = (c) => this.applyToContext(objects, c, config, response)
    for (const isemantic in this.semantics) {
      const semantic = this.semantics[isemantic]
      if (semantic.matches(objects, context, config)) {
        context_prime = semantic.apply(objects, context, config, response, s);
        this.logs.push(`Semantics: applied ${semantic.toString()}\n to\n ${JSON.stringify(context)}\n result is ${JSON.stringify(context_prime)}`)
        break
      }
    }
    return context_prime
  }

  applyToContexts(objects, contexts, config) {
    const contexts_prime = [];
    contexts.forEach( (context) => {
      contexts_prime.push(this.applyToContext(objects, context, config));
    });
    return contexts_prime;
  }

  apply(objects, value, config, response) {
    if ( Array.isArray(value) ) {
      return this.applyToContexts(objects, value, config, response)
    } else if ( value instanceof Object) {
      return this.applyToContext(objects, value, config, response)
    } else {
      return value
    }
  }
}

module.exports = { Semantic, Semantics };
